#include	"ChunkHeader.h"
#include	"FmtHeader.h"
#include	"RiffHeader.h"
#include	"WavRegion.h"

#include	<iostream>
#include	<fstream>
#include	<sstream>
#include	<string>



WavRegion::WavRegion (const std::string &		source_filename,
                      const std::streampos	source_data_pos,
                      const FmtHeader &		fmt)
:	_source_filename (source_filename)
,	_source_data_pos (source_data_pos)
,	_fmt (fmt)
,	_beg (0)
,	_end (0)
,	_number (0)
,	_amp (1)
{
}



void	WavRegion::set_start (long beg)
{
	_beg = beg;
}



void	WavRegion::set_end (long end)
{
	_end = end;
}



void	WavRegion::set_number (int number)
{
	_number = number;
}



void	WavRegion::set_amp (int amp)
{
	if (amp < 1)
	{
		amp = 1;
	}
	_amp = amp;
}


int	WavRegion::save (std::istream &source_file, bool mono_flag) const
{
	using namespace std;

	/* New format */
	FmtHeader	region_fmt = _fmt;
	if (mono_flag)
	{
		region_fmt._channels = 1;
		region_fmt._block_align = (region_fmt._bits_per_sample + 7) >> 3;
	}

	const long	dot_pos = _source_filename.rfind ('.');
	string		filename = _source_filename.substr (0, dot_pos);
	string		ext;

	if (dot_pos != string::npos)
	{
		ext = _source_filename.substr (dot_pos, string::npos);
	}

	string	number_s;
	stringstream	number_ss;
	number_ss.fill ('0');
	number_ss.width (2);
	number_ss << _number;
	number_ss >> number_s;

	filename += "-" + number_s + ext;

	clog << "Saving region: " << filename
	     << "\n\tFrom     : " << _beg
	     << "\n\tLen      : " << (_end - _beg)
	     << "\n\tChannels : " << region_fmt._channels
	     << "\n\tAmplify  : " << _amp
	     << '\n';

	ofstream	region_file (filename.c_str (), ios::binary);
	if (! region_file)
	{
		cerr << "Can't open " << filename << " for writing.\n";
		return (-1);
	}

	const long	data_len = _end - _beg;
	const long	data_len_char = data_len * region_fmt._block_align;
	ChunkHeader	hdr;
	int	res = 0;


/*______________________________________________
 *
 * File header
 *______________________________________________
 */

	hdr._size =   RiffHeader::get_len ()
	            + ChunkHeader::get_len () + FmtHeader::get_len ()
	            + ChunkHeader::get_len () + data_len_char;
	hdr.set ("RIFF");
	RiffHeader	riff;
	riff.set ("WAVE");

	res |= hdr.write (region_file);
	res |= riff.write (region_file);

	if (res != 0)
	{
		cerr << "Can't write region file header.\n";
		return (-1);
	}

/*______________________________________________
 *
 * Fmt
 *______________________________________________
 */

	hdr._size = FmtHeader::get_len ();
	hdr.set ("fmt ");

	res |= hdr.write (region_file);
	res |= region_fmt.write (region_file);

	if (res != 0)
	{
		cerr << "Can't write region fmt chunk.\n";
		return (-1);
	}

/*______________________________________________
 *
 * Data
 *______________________________________________
 */

	hdr._size = data_len_char;
	hdr.set ("data");

	if (hdr.write (region_file))
	{
		cerr << "Can't write region data chunk header.\n";
		return (-1);
	}

	/* Save input file position */
	streampos	old_source_pos = source_file.tellg ();
	if (old_source_pos == streampos (-1))
	{
		cerr << "Can't save source file position.\n";
		return (-1);
	}

	/* Go to region begining in the source file */
	streampos	region_pos = _source_data_pos;
	region_pos += _beg * _fmt._block_align;
	if (! source_file.seekg (region_pos))
	{
		cerr << "Can't seek region begining.\n";
		return (-1);
	}

	/* Copy */
	Int16		buffer;
	for (long cur_pos = 0; cur_pos < data_len; ++ cur_pos)
	{
		for (int chn = 0; chn < _fmt._channels; ++ chn)
		{
			res |= read_le (source_file, buffer);
			if (chn < region_fmt._channels)
			{
				buffer *= _amp;
				res |= write_le (region_file, buffer);
			}
		}

		if (res != 0)
		{
			cerr << "Can't write region data at position " << cur_pos << ".\n";
			return (-1);
		}
	}

	/* Restore input file position */
	if (! source_file.seekg (old_source_pos))
	{
		cerr << "Can't restore source file position.\n";
		return (-1);
	}

/*______________________________________________
 *
 * Close file
 *______________________________________________
 */

	region_file.close ();
	if (! region_file)
	{
		cerr << "Can't close " << filename << ".\n";
		return (-1);
	}

	return (0);
}
