/*

Splits large WAV files containting pieces of sounds separated by silences.
This is intended to spit out usable sound files from audio sample-CD tracks.

(c) 1999 - Laurent de Soras (ldesoras@club-internet.fr)

*/

#include	"ChunkHeader.h"
#include	"FmtHeader.h"
#include	"RiffHeader.h"
#include	"WavRegion.h"

#include	<iostream>
#include	<fstream>
#include	<string>
#include	<vector>


const	int	HEADER_FMT  = 0x01;
const int	HEADER_DATA = 0x02;

const double	min_silence_duration = 0.03;	// seconds
const double	min_region_duration = 0.1;		// seconds
const Int16		attack_thresold = 10;
const Int16		release_thresold = 10;



static void	print_syntax ();



int	main (int argc, char *argv [])
{
	using namespace std;

/*______________________________________________
 *
 * Read arguments
 *______________________________________________
 */

	bool				auto_normalization_flag = true;

	if (argc < 2)
	{
		print_syntax ();
		return (-1);
	}

	int				cur_opt = 1;
	while (cur_opt < argc - 1)
	{
		const std::string		opt_name (argv [cur_opt]);
		++cur_opt;

		if (opt_name == "--normalize")
		{
			if (cur_opt >= argc - 1)
			{
				print_syntax ();
				return (-1);
			}

			const std::string	opt_val (argv [cur_opt]);
			++cur_opt;

			auto_normalization_flag = (opt_val == "on");
		}

		else
		{
			print_syntax ();
			return (-1);
		}
	}

	string		main_filename (argv [argc - 1]);
	clog << "Input file: \"" << main_filename << "\"\n";

	ifstream	main_file (main_filename.c_str (), ios::binary);
	if (! main_file)
	{
		cerr << "Can't open " << main_filename << " for reading.\n";
		return (-1);
	}

	ChunkHeader		hdr;
	FmtHeader		fmt;

/*______________________________________________
 *
 * File header
 *______________________________________________
 */

	RiffHeader		riff;

	if (hdr.read (main_file))
	{
		cerr << "Can't read file header.\n";
		return (-1);
	}

	if (! hdr.cmp ("RIFF"))
	{
		cerr << "Not a RIFF file.\n";
		return (-1);
	}

	if (riff.read (main_file))
	{
		cerr << "Can't read file header.\n";
		return (-1);
	}

	if (! riff.is_wave ())
	{
		cerr << "Not a WAV file.\n";
		return (-1);
	}

/*______________________________________________
 *
 * Chunk headers
 *______________________________________________
 */

	streampos	data_pos;
	long			data_len_char;
	int			flags = 0;

	while (flags != (HEADER_FMT | HEADER_DATA))
	{
      if (hdr.read (main_file))
		{
			if (main_file.eof ())
			{
				cerr << "Unexpected end of file (mandatory chunks needed).\n";
			}
			else
			{
				cerr << "Can't read chunk header.\n";
			}
			return (-1);
		}

		if (hdr.cmp ("fmt "))
		{
			if (fmt.read (main_file))
			{
				cerr << "Can't read fmt chunk.\n";
				return (-1);
			}

			if (fmt._format_tag != 1)
			{
				cerr << "Only PCM format is supported.\n";
				return (-1);
			}

			if (fmt._bits_per_sample > 16)
			{
				cerr << "Sample resolution higher than 16 bit is not supported.\n";
				return (-1);
			}

			const int	char_width = (fmt._bits_per_sample + 7) >> 3;
			if (char_width * fmt._channels != fmt._block_align)
			{
				cerr << "Samples are not contiguous.\n";
				return (-1);
			}

			flags |= HEADER_FMT;
		}

		else if (hdr.cmp ("data"))
		{
			data_len_char = hdr._size;
			data_pos = main_file.tellg ();
			flags |= HEADER_DATA;
		}

		hdr.skip (main_file);
	}

/*______________________________________________
 *
 * Sample processing
 *______________________________________________
 */

	if (! main_file.seekg (data_pos))
	{
		cerr << "Can't reach data begining.\n";
		return (-1);
	}

	const long	data_len = data_len_char / fmt._block_align;
	const long	min_silence_len = min_silence_duration * fmt._samples_per_sec;
	const long	min_region_len  = min_region_duration * fmt._samples_per_sec;

	long				region_beg = -1;
	long				region_end = -1;
	int				region_nbr = 0;
	bool				region_mono;
	double			region_peak;
	WavRegion		region (main_filename, data_pos, fmt);
	vector <Int16>	data (fmt._channels);

	long	cur_pos = 0;
	while (cur_pos < data_len)
	{
		bool		attack = false;
		bool		release = false;

		for (int chn = 0; chn < fmt._channels; ++chn)
		{
			if (read_le (main_file, data [chn]))
			{
				cerr << "Can't read data (sample "
				     << cur_pos << ").\n";
				return (-1);
			}

			if (abs (data [chn]) > attack_thresold)
			{
				attack = true;
			}

			if (abs (data [chn]) <= release_thresold)
			{
				release = true;
			}
		}

		// Detect region begining
		if (region_beg < 0)
		{
			if (attack)
			{
				region_beg = cur_pos;
				region_mono = true;
				region_peak = 0;
			}
		}

		else
		{
			// Detect stereo
			if (region_mono)
			{
				for (chn = 1; chn < fmt._channels; ++chn)
				{
					if (data [chn] != data [0])
					{
						region_mono = false;
					}
				}
			}

			// Detect peak
			for (chn = 0; chn < fmt._channels; ++chn)
			{
				const double	instant_data = abs (double (data [chn]));
				if (instant_data > region_peak)
				{
					region_peak = instant_data;
				}
			}

			// Detect region end
			if (release)
			{
				if (region_end < 0)
				{
					region_end = cur_pos;
				}
			}

			else
			{
				region_end = -1;
			}
		}

		// Next position
		++ cur_pos;

		bool		save_region = false;

		if (   region_end >= 0
		    && cur_pos - min_silence_len > region_end)
		{
			save_region = true;
		}
		else if (   cur_pos >= data_len
		         && region_beg >= 0)
		{
			if (region_end < 0)
			{
				region_end = cur_pos;
			}
			save_region = true;
		}

		// Validate region end
		if (save_region)
		{
			// Check whether region is long enough
			if (region_end - region_beg > min_region_len)
			{
				region.set_start (region_beg);
				region.set_end (region_end);
				region.set_number (region_nbr);

				// Normalize
				if (auto_normalization_flag && region_peak > 0)
				{
					region.set_amp (int (32767.0 / region_peak));
				}

				if (region.save (main_file, region_mono))
				{
					return (-1);
				}

				++ region_nbr;
			}

			region_beg = -1;
			region_end = -1;
		}
	}



/*______________________________________________
 *
 * Close file
 *______________________________________________
 */

	main_file.close ();
	clog << "Successfully done.\n";

	return (0);
}



static void	print_syntax ()
{
	using namespace std;

	cerr << "Syntax:\n"
	     << "WavSplit <options> <filename>\n";
	cerr << "Options:\n"
	     << "--normalize [on|off]: Turn on or off auto-normalization. Default is On.\n";
	cerr << "\n";
}
